/// ***************************************************************************
/// ***************************************************************************
/// unit name: sboolsearch.pas
/// author: mrbeeye, mrbeeye(at)safe-mail.net
/// project: Simple Boolean Search
/// unit version: 0.1
/// project version: 0.1
/// creation date&time stamp: 05/03/2005, 2:24 PM
/// ---------------------------------------------------------------------------
/// unit license: free
/// project license: free
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
/// ---------------------------------------------------------------------------
/// note:
/// Based on unit BoolPos (version 1.3.5.P) by Bruce J. Lackore.
/// Unit may be found on Borland Developer Network (keywords: +boolean +search)
/// in Code Central
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// legend: + added feature; - removed feature; * bugfix; ^ improvement
///
/// mrbeeye: v0.1: unit is just transcript of BoolPos into Delphi
///
/// ***************************************************************************
/// ***************************************************************************
unit sboolsearch;

interface

type
  TSimpleBooleanSearch = class
    private
      FPattern,
      FInput: AnsiString;

      function CountChar(const Input: AnsiString; const Char: AnsiChar): Word;
      function PosChar(const Input: AnsiString; const Char: AnsiChar): Boolean;
      function LastCharPos(const Input: AnsiString; const Char: AnsiChar): Word;
      function NextCharPos(const Input: AnsiString; const Char: AnsiChar; const FromIndex: Word): Word;
      function FixPattern(Pattern: AnsiString): AnsiString;
      function ParsePattern(Pattern, Input: AnsiString): AnsiString;
      function ProcessTokenStr(Token: AnsiString): AnsiChar;
      function BPos(Input, Pattern: AnsiString; const IgnoreCase: Boolean):  Boolean;
    public
      constructor Create;
      destructor Free;

      procedure SetPattern(const Pattern: AnsiString);
      procedure SetInput(const Input: AnsiString);
      function GetPattern: AnsiString;
      function GetInput: AnsiString;

      function Search(const CaseSensitivity: Boolean): Boolean; overload;
      function Search(const Input, Pattern: AnsiString; const CaseSensitivity: Boolean): Boolean; overload;
  end;

implementation

uses
  SysUtils, StrUtils;

const
  NULL_STRING:        AnsiString  = '';

  LEFT_PARENTHESIS:   Char        = '(';
  RIGHT_PARENTHESIS:  Char        = ')';

  OPERATOR_AND:       Char        = '&';
  OPERATOR_OR:        Char        = '|';

  TOKEN_TRUE:         Char        = 'T';
  TOKEN_FALSE:        Char        = 'F';

/// ***************************************************************************
/// name: Create (constructor)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   -none-
///
/// output:
///   -none-
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Create TSimpleBooleanSearch instance
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
constructor TSimpleBooleanSearch.Create;
begin
  Self.FInput   := NULL_STRING;
  Self.FPattern := NULL_STRING;
end;

/// ***************************************************************************
/// name: Create (destructor)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   -none-
///
/// output:
///   -none-
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Free TSimpleBooleanSearch instance
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
destructor TSimpleBooleanSearch.Free;
begin
end;

/// ***************************************************************************
/// name: CountChar (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Char - character, which will be searched for
///
/// output:
///   Number of occurences of Char in Input
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return number of Char in Input
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.CountChar(const Input: AnsiString; const Char: AnsiChar): Word;
var
  i, j: Word;
begin
  j := 0;

  for i := 1 to Length(Input) do
    if (Input[i] = Char) then
      Inc(j);

  Result := j;
end;

/// ***************************************************************************
/// name: PosChar (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Char - character, which will be searched for
///
/// output:
///   TRUE if Char found "in the middle" of Input, otherwise FALSE
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return TRUE if Char is found in Input and Char is not first or last
///   character in INput 
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.PosChar(const Input: AnsiString; const Char: AnsiChar): Boolean;
var
  j: Word;
begin
  j := PosEx(Char, Input, 1);

  Result := ( (j > 0) and (j < Length(Input)) );
end;

/// ***************************************************************************
/// name: LastCharPos (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Char - character, which will be searched for
///
/// output:
///   Position of last occurence of Char in Input. If Char not found return 0.
///   Zero (0) if Char not found.
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return last position of Char in Input.
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.LastCharPos(const Input: AnsiString;
  const Char: AnsiChar): Word;
var
  i, j: Word;
begin
  j := 0;

  for i := 0 to Length(Input) do
    if (Input[i] = Char) then j := i;

  Result := j;
end;

/// ***************************************************************************
/// name: NextCharPos (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Char - character, which will be searched for
///
/// output:
///   Position of next occurence of Char in Input from FromIndex position.
///   Zero (0) if Char not found.
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return next position of Char in Input from next char from FromIndex position
/// ---------------------------------------------------------------------------
/// note:
///   Return the offset from the beginning of Input
///   Return next position of Char in Input AFTER (not including) FromIndex
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.NextCharPos(const Input: AnsiString; const Char: AnsiChar;
 const FromIndex: Word): Word;
begin
  Result := PosEx(Char, Input, (FromIndex + 1));
end;

/// ***************************************************************************
/// name: FixPattern (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Pattern - pattern string
///
/// output:
///   Fixed pattern if everything OK otherwise empty string
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Count the number of paranthesis pairs and correct for a deficient number of
///   either by adding the appropriate character either at the beginning or
///   the end of the search string.
/// ---------------------------------------------------------------------------
/// note:
///   It will add one set of parantheses if none are found.
///   Function also looks for illegal character pairs (&, &), (| and |),
///   these pairs indicate an illegal boolean search.
///   The function returns the corrected Patterm if all is well,
///   an empty string if not.
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.FixPattern(Pattern: AnsiString): AnsiString;
var
  LeftPar, RightPar, Count: Integer;
begin
  if ((Pattern[Length(Pattern)] <> RIGHT_PARENTHESIS) or
     (Pattern[1] <> LEFT_PARENTHESIS)) then
     Pattern := LEFT_PARENTHESIS + Pattern + RIGHT_PARENTHESIS;

  LeftPar   := CountChar(Pattern, LEFT_PARENTHESIS);
  RightPar  := CountChar(Pattern, RIGHT_PARENTHESIS);

  Count := Abs(LeftPar - RightPar);

  if (Count > 0) then
    if (RightPar < LeftPar) then
      Pattern := Pattern + DupeString(RIGHT_PARENTHESIS, Count)
    else
      Pattern := DupeString(LEFT_PARENTHESIS, Count) + Pattern;

  if (Pos(LEFT_PARENTHESIS + OPERATOR_AND, Pattern) > 0)  or
     (Pos(OPERATOR_AND + RIGHT_PARENTHESIS, Pattern) > 0) or
     (Pos(LEFT_PARENTHESIS + OPERATOR_OR, Pattern) > 0)   or
     (Pos(OPERATOR_OR + RIGHT_PARENTHESIS, Pattern) > 0) then
    Result := NULL_STRING
  else
    Result := Pattern;
end;

/// ***************************************************************************
/// name: ParsePattern (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Pattern - pattern string
///   Input - input (source) string
///
/// output:
///   Tokenized pattern
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Extract each string to search for, test to see if it exists in the original
///   string and replaces the extracted substring with the appropriate token.
/// ---------------------------------------------------------------------------
/// note:
///   Each substring is determined solely by the characters used for paranthesis
///   and operators.  Any other characters are assumed to be part of the search
///   string.
///   Each substring is searched for in the Pattern and its presense or absense
///   noted with a T or F respectively.
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.ParsePattern(Pattern, Input: AnsiString): AnsiString;
var
  RetString, TokenString: AnsiString;
  EndToken: Boolean;
begin
  TokenString := NULL_STRING;
  RetString   := NULL_STRING;

  while (Pattern <> NULL_STRING) do
    begin
      if (Pattern[1] in [LEFT_PARENTHESIS, RIGHT_PARENTHESIS, OPERATOR_AND, OPERATOR_OR]) then
        begin
          EndToken := (TokenString <> NULL_STRING);

          if (EndToken = FALSE) then
            RetString := RetString + Pattern[1]
        end
      else
        begin
          TokenString := TokenString + Pattern[1];
          EndToken := FALSE;
        end;

      if (EndToken = TRUE) then
        begin
          if (Pos(TokenString, Input) > 0) then
            RetString := RetString + TOKEN_TRUE
          else
            RetString := RetString + TOKEN_FALSE;

          RetString   := RetString + Pattern[1];
          TokenString := NULL_STRING;

          EndToken := FALSE;
        end;

      Delete(Pattern, 1, 1);
    end;

    Result := RetString;
end;

/// ***************************************************************************
/// name: ProcessTokenStr (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Token - substring
///
/// output:
///   Token TRUE or token FALSE
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return final boolean token (TOKEN_TRUE or TOKEN_FALSE)
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.ProcessTokenStr(Token: AnsiString): AnsiChar;
var
  One_token: AnsiString;
  One_token_len, LParenthesis, j: Word;

  function ProcessAToken(AToken: AnsiString): AnsiChar;
  var
    Lcv: Word;
    CurrentAnswer, DoAND:  Boolean;
  begin
    CurrentAnswer := (AToken[1] = TOKEN_TRUE);

    DoAND := FALSE;

    for Lcv := 2 to Length(AToken) do
      if (AToken[Lcv] = OPERATOR_AND) then
        DoAND := TRUE
      else
        if (AToken[Lcv] = OPERATOR_OR) then
          DoAND := FALSE
        else
          if (AToken[Lcv] = TOKEN_TRUE) then
            if (DoAND = TRUE) then
              CurrentAnswer := (CurrentAnswer and TRUE)
            else
              CurrentAnswer := TRUE
          else
            if (AToken[Lcv] = TOKEN_FALSE) then
              if (DoAND = TRUE) then
                CurrentAnswer := FALSE;

    if (CurrentAnswer = TRUE) then
      Result := TOKEN_TRUE
    else
      Result := TOKEN_FALSE;
  end;

begin
  if PosChar(Token, LEFT_PARENTHESIS) then
    begin
      j := Length(Token);

      while (j > 1) do
        begin
          if (PosChar(Token, LEFT_PARENTHESIS) = FALSE) then
            Token := LEFT_PARENTHESIS + Token + RIGHT_PARENTHESIS;

          LParenthesis := LastCharPos(Token, LEFT_PARENTHESIS);

          One_token_len := Succ(NextCharPos(Token, RIGHT_PARENTHESIS, LParenthesis) - LParenthesis);

          One_token := Copy(Token, LParenthesis, One_token_len);

          Dec(j);
          SetLength(Token, j);
          Delete(One_token, 1, 1);

          Delete(Token, LParenthesis, One_token_len);

          Insert(ProcessAToken(One_token), Token, LParenthesis)
        end;

      Result := Token[1];
    end
  else
    Result := ProcessAToken(One_token);
end;

/// ***************************************************************************
/// name: ProcessTokenStr (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Pattern - pattern string which will be searched for in Input
///   CaseSensitivity - TRUE if case sensitive, otherwise FALSE
///
/// output:
///   TRUE if Pattern found in Input
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.BPos(Input, Pattern: AnsiString;
 const IgnoreCase: Boolean): Boolean;
begin
  Result := FALSE;

  if (IgnoreCase = TRUE) then
    begin
      Pattern := UpperCase(Pattern);
      Input   := UpperCase(Input);
    end;

  if ((PosChar(Pattern, OPERATOR_OR)  = TRUE) or
     (PosChar(Pattern, OPERATOR_AND) = TRUE)) then
    begin
      Pattern := ParsePattern(FixPattern(Pattern), Input);

      if (Pattern <> NULL_STRING) then
        Result := (ProcessTokenStr(Pattern) = TOKEN_TRUE)
    end
  else
    Result := AnsiContainsText(Input, Pattern);
end;

/// ***************************************************************************
/// name: SetPattern (procedure)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Pattern - pattern string
///
/// output:
///   -none-
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Set pattern string
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
procedure TSimpleBooleanSearch.SetPattern(const Pattern: AnsiString);
begin
  Self.FPattern := Pattern;
end;

/// ***************************************************************************
/// name: SetInput (procedure)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///
/// output:
///   -none-
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Set input (source) string
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
procedure TSimpleBooleanSearch.SetInput(const Input: AnsiString);
begin
  Self.FInput := Input;
end;

/// ***************************************************************************
/// name: GetInput (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   -none-
///
/// output:
///   Input
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return input (source) string
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.GetInput: AnsiString;
begin
  Result := Self.FInput;
end;

/// ***************************************************************************
/// name: Search (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   -none-
///
/// output:
///   Pattern
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
///   Return pattern string
/// ---------------------------------------------------------------------------
/// note:
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.GetPattern: AnsiString;
begin
  Result := Self.FPattern;
end;

/// ***************************************************************************
/// name: Search (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   CaseSensitivity - TRUE if case sensitive, otherwise FALSE
///
/// output:
///   TRUE if Pattern found in Input, otherwise FALSE
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
/// ---------------------------------------------------------------------------
/// note:
///   overloaded function
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.Search(const CaseSensitivity: Boolean): Boolean;
begin
  Result := FALSE;

  if ((Self.FInput <> NULL_STRING) and (Self.FPattern <> NULL_STRING)) then
    Result := BPos(Self.FInput, Self.FPattern, CaseSensitivity);
end;

/// ***************************************************************************
/// name: Search (function)
/// scope: TSimpleBooleanSearch
/// version: 0.1
/// author: mrbeeye
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// input:
///   Input - input (source) string
///   Pattern - pattern string which will be searched for in Input
///   CaseSensitivity - TRUE if case sensitive, otherwise FALSE
///
/// output:
///   TRUE if Pattern found in Input, otherwise FALSE
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// description:
/// ---------------------------------------------------------------------------
/// note:
///   overloaded function
/// ---------------------------------------------------------------------------
/// ---------------------------------------------------------------------------
/// changelog:
/// ***************************************************************************
function TSimpleBooleanSearch.Search(const Input, Pattern: AnsiString;
 const CaseSensitivity: Boolean): Boolean;
begin
  Result := FALSE;

  Self.FInput   := Input;
  Self.FPattern := Pattern;

  Result := Self.Search(CaseSensitivity);
end;

end.
